/********************************************************/
/* Vex Decoder.c                                        */
/*                                                      */
/* Converts the serial output from the VEX receiver     */
/* into six values readable by I2C.                     */
/*    four 8-bit values (analog)                        */
/*    two 1-bit values (digital)                        */
/*                                                      */
/* Kerry C. Imming                                      */
/* kcimming@magnaspeed.net                              */
/* October 18, 2008                                     */
/*                                                      */
/* Copyright (c) 2008 Kerry Imming                      */
/* Some Rights Reserved                                 */
/* -- see http://creativecommons.org/licenses/by/3.0/   */
/*                                                      */
/********************************************************/

// Slave I2C Interrupt handler based on Microchip Application Note AN734
// by Stephen Bowling and Naveen Raj, January 2008
// 
// VEX receiver decoding algorithm based on Jon Williams' "VEX Receiver
// Decoder Circuit/Firmware", the header of which is included here:
//' =========================================================================
//'
//' File...... VEX_Demod-v2.SXB
//' Purpose... Servo demodulator for VEX receiver
//' Author.... Jon Williams
//' Copyright (c) 2007 Jon Williams
//' Some Rights Reserved
//' -- see http://creativecommons.org/licenses/by/2.5/
//' E-mail.... jwilliams@efx-tek.com
//' Started...
//' Updated... 02 MAR 2007
//'
//' =========================================================================

#include <htc.h>

#define BYTE unsigned char
#define ISON 0xFF
#define ISOFF 0x00

/* I2C address of this node 64 decimal */
#define NODE_ADDR       0x40
#define NODE_MASK       0xF0


/* Program device configuration word
 * Oscillator = Internal, I/O function on clock pins
 * Watchdog Timer = On
 * Power Up Timer = Off
 * Master Clear Enable = Internal
 * Code Protect = Off
 * Data EE Read Protect = Off
 * Brown Out Detect = BOD and SBOREN disabled
 * Internal External Switch Over Mode = Disabled
 * Monitor Clock Fail-safe = Disabled
 */
__CONFIG(INTIO & WDTEN & PWRTDIS & MCLRDIS & UNPROTECT & UNPROTECT & BORDIS & IESODIS & FCMDIS);

  static BYTE   chdata[8];    /* 6 used, reserve 4C & 4E to simplify masking */
  static BYTE   i2c_address;
  static BYTE   i2c_err;
 

void WriteI2C(BYTE data) {
  while (BF) { }

  do {
     WCOL = 0;
     SSPBUF = data;
  }
  while (WCOL == 1);

  CKP = 1;   /* Release the clock. */
  return;
}        

void SSP_Handler(void) {
  BYTE state;
  static BYTE read_cmd;

   state = SSPSTAT & 0b00101101;        // capture masked version of SSPSTAT for state checking

   /* State1:  Write operation, last byte was an address, buffer is full. */
   if (state == 0b00001001) {
      i2c_address = (SSPBUF >> 1) & 0x07;
//    PORTC = 0x01;
      return;
   }
   /* State2: Write operation, last byte was data, buffer is full. */
   else if (state == 0b00101001) {
      chdata[i2c_address] = SSPBUF;
      i2c_address = i2c_address++ & 0x07;
//    PORTC = 0x02;
      return;
   }
   /* State3: First Read operation, last byte was an address, buffer is empty. */
   else if (state == 0b00001100) {
      i2c_address = (SSPBUF >> 1) & 0x07;
      WriteI2C(chdata[i2c_address]);
      i2c_address = i2c_address++ & 0x07;
      read_cmd = 1;     
//    PORTC = 0x04;
      return;
   }
   /* State4: Other reads operation, last byte was data, buffer is empty. */
   else if (state == 0b00101100) {
      WriteI2C(chdata[i2c_address]);
      i2c_address = i2c_address++ & 0x07;     
//    PORTC = 0x08;
      return;
   }
   /* State5:  A NACK was received when transmitting data back from the master.
               Slave logic is reset in this case  */
   /* Last read operation,    buffer is empty       */
   else if (state == 0b00101000) {
      if (read_cmd) {
         read_cmd = 0;
      }
      else {
         /* I2C Error */
//       while(1) {
//          PORTC = 0x01;
//          PORTC = 0x02;
//          PORTC = 0x04;
//          PORTC = 0x08;
//       }
      }
      return;
   }
}

static void interrupt isr(void)
{
        
  if(!SSPIF) {               // Was this an I2C interrupt?
     chdata[7]++;            // no, set status flag
     RC3 = 1;
     while (1) {}            // HANG!  wait for watchdog timer to reset chip
  }
  SSPIF = 0;                 // Clear interrupt flag, ready for next
  SSP_Handler();
                        
//PORTC = PORTC ^ 0x08;      // Toggle PortC, bit 3        
//PORTC = PORTC ^ 0x08; 
                             // Return from interrupt re-enables global interrupts
}


void main(void)
{

   /* Input pin is RA2, with external 10K pull-up */
   /* RB4 and RB6 are used by SSP, external 82K pull-up required  */
 
   BYTE pulseTmr;
   BYTE ch;        
   unsigned int syncTmr;

   WDTCON  = 0x11;           // 8k pre-scalar, watchdog enabled
   OSCCON  = 0b01100000;     // Internal oscillator set to 4MHz

   TRISC = 0xF0;             // 7:4 = input, 3:0 = output
   TRISA = 0xFF;             // All input

   ADCON1 = 0x00;            // NOT: A2D clock Fosc/8
   ANSEL  = 0x00;            // PORTA all digital
   ADCON0 = 0x00;            // NOT: select channel 1 on A2D

   POR = 1;
   BOR = 1;

   PORTB = 0;                // Clear PORTB
 
   PIR1 = 0;
   TRISB = 0xFF;             // Port B all input
   ANSELH = 0;               // Port B all digital

   SSPCON = 0x39;            // SSP Errata work around
   SSPADD = NODE_MASK;       // set Node Mask 
   SSPCON = 0x36;            // Setup SSP module for 7-bit address, slave mode
   SSPADD = NODE_ADDR;       // set Node Address
   SSPSTAT = 0;
    
   SSPIE = 1; 
   PEIE = 1;                 // Enable all peripheral interrupts
   GIE = 1;                  // Enable global interrupts

   PORTC = 0;
// RC3 = 1;                  // Set PORTC bit 3, DSP4


  while (1) {
          
//   RC0 = RC0 ^ 1;
//   PORTC = chdata[1];         // copy register 1 on PORTC

     /* Wait For Sync */  
     syncTmr = 0;
     while (syncTmr < 1000) {
        syncTmr = 1;
        while (RA2 == 0) {
	   syncTmr++;
	}
     }
     chdata[7] = syncTmr >> 8;

     /* Analog Channels */
     for (ch=0; ch<4; ch++) {
        pulseTmr = 0;
        while (RA2 == 1) {          /* complete framing pulse */
	} 
	while (RA2 == 0) {          /* measure timing pulse */
	   asm("nop");
           asm("nop");
           asm("nop");
           pulseTmr++;
	}
        chdata[ch] = pulseTmr;
     } 
       
     /* Control Channels */
     for (ch=4; ch<6; ch++) {
        pulseTmr = 0;
        while (RA2 == 1) {          /* complete framing pulse */
        }
        while (RA2 == 0) {          /* measure timing pulse */
           asm("nop");
           asm("nop");
           asm("nop");
           pulseTmr++;
        }
       if (pulseTmr < 120)          /* Short Pulse? */
          chdata[ch] = ISON;
       else if (pulseTmr > 170)     /* Long Pulse? */
          chdata[ch] = ISOFF;
     }
  }

}
